package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;

public class DoStatement extends Statement {
  public Expression condition;
  
  public Statement action;
  
  private BranchLabel breakLabel;
  
  private BranchLabel continueLabel;
  
  int mergedInitStateIndex = -1;
  
  int preConditionInitStateIndex = -1;
  
  public DoStatement(Expression condition, Statement action, int sourceStart, int sourceEnd) {
    this.sourceStart = sourceStart;
    this.sourceEnd = sourceEnd;
    this.condition = condition;
    this.action = action;
    if (action instanceof org.eclipse.jdt.internal.compiler.ast.EmptyStatement)
      action.bits |= 0x1; 
  }
  
  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    this.breakLabel = new BranchLabel();
    this.continueLabel = new BranchLabel();
    LoopingFlowContext loopingContext = 
      new LoopingFlowContext(
        flowContext, 
        flowInfo, 
        (ASTNode)this, 
        this.breakLabel, 
        this.continueLabel, 
        (Scope)currentScope, 
        false);
    Constant cst = this.condition.constant;
    boolean isConditionTrue = (cst != Constant.NotAConstant && cst.booleanValue());
    cst = this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedTrue = (cst != Constant.NotAConstant && cst.booleanValue());
    boolean isConditionOptimizedFalse = (cst != Constant.NotAConstant && !cst.booleanValue());
    int previousMode = flowInfo.reachMode();
    FlowInfo initsOnCondition = flowInfo;
    UnconditionalFlowInfo actionInfo = flowInfo.nullInfoLessUnconditionalCopy();
    if (this.action != null && !this.action.isEmptyBlock()) {
      actionInfo = this.action
        .analyseCode(currentScope, (FlowContext)loopingContext, (FlowInfo)actionInfo)
        .unconditionalInits();
      if ((actionInfo.tagBits & 
        loopingContext.initsOnContinue.tagBits & 
        0x1) != 0)
        this.continueLabel = null; 
      if ((this.condition.implicitConversion & 0x400) != 0)
        initsOnCondition = flowInfo.unconditionalInits()
          .addInitializationsFrom(
            (FlowInfo)actionInfo.mergedWith(loopingContext.initsOnContinue)); 
    } 
    this.condition.checkNPEbyUnboxing(currentScope, flowContext, initsOnCondition);
    actionInfo.setReachMode(previousMode);
    LoopingFlowContext condLoopContext;
    FlowInfo condInfo = 
      this.condition.analyseCode(
        currentScope, 
        (FlowContext)(condLoopContext = 
        new LoopingFlowContext(flowContext, flowInfo, (ASTNode)this, null, 
          null, (Scope)currentScope, true)), (
        (this.action == null) ? 
        actionInfo : 
        actionInfo.mergedWith(loopingContext.initsOnContinue)).copy());
    this.preConditionInitStateIndex = currentScope.methodScope().recordInitializationStates((FlowInfo)actionInfo.mergedWith(loopingContext.initsOnContinue));
    if (!isConditionOptimizedFalse && this.continueLabel != null) {
      loopingContext.complainOnDeferredFinalChecks(currentScope, condInfo);
      condLoopContext.complainOnDeferredFinalChecks(currentScope, condInfo);
      loopingContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)flowInfo.unconditionalCopy().addPotentialNullInfoFrom(
            condInfo.initsWhenTrue().unconditionalInits()));
      condLoopContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)actionInfo.addPotentialNullInfoFrom(
            condInfo.initsWhenTrue().unconditionalInits()));
    } else {
      loopingContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)flowInfo.unconditionalCopy().addPotentialNullInfoFrom(
            condInfo.initsWhenTrue().unconditionalInits()), false);
      condLoopContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)actionInfo.addPotentialNullInfoFrom(
            condInfo.initsWhenTrue().unconditionalInits()), false);
    } 
    if (loopingContext.hasEscapingExceptions()) {
      FlowInfo loopbackFlowInfo = flowInfo.copy();
      UnconditionalFlowInfo unconditionalFlowInfo = loopbackFlowInfo.mergedWith(loopbackFlowInfo.unconditionalCopy().addNullInfoFrom(condInfo.initsWhenTrue()).unconditionalInits());
      loopingContext.simulateThrowAfterLoopBack((FlowInfo)unconditionalFlowInfo);
    } 
    UnconditionalFlowInfo unconditionalFlowInfo1 = 
      FlowInfo.mergedOptimizedBranches(
        ((loopingContext.initsOnBreak.tagBits & 0x3) != 0) ? 
        (FlowInfo)loopingContext.initsOnBreak : 
        flowInfo.unconditionalCopy().addInitializationsFrom((FlowInfo)loopingContext.initsOnBreak), 
        
        isConditionOptimizedTrue, 
        ((condInfo.tagBits & 0x3) == 0) ? 
        flowInfo.copy().addInitializationsFrom(condInfo.initsWhenFalse()) : 
        condInfo, 
        
        false, 
        !isConditionTrue);
    this.mergedInitStateIndex = currentScope.methodScope().recordInitializationStates((FlowInfo)unconditionalFlowInfo1);
    return (FlowInfo)unconditionalFlowInfo1;
  }
  
  public void generateCode(BlockScope currentScope, CodeStream codeStream) {
    if ((this.bits & Integer.MIN_VALUE) == 0)
      return; 
    int pc = codeStream.position;
    BranchLabel actionLabel = new BranchLabel(codeStream);
    if (this.action != null)
      actionLabel.tagBits |= 0x2; 
    actionLabel.place();
    this.breakLabel.initialize(codeStream);
    boolean hasContinueLabel = (this.continueLabel != null);
    if (hasContinueLabel)
      this.continueLabel.initialize(codeStream); 
    if (this.action != null)
      this.action.generateCode(currentScope, codeStream); 
    if (hasContinueLabel) {
      this.continueLabel.place();
      if (this.preConditionInitStateIndex != -1) {
        codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.preConditionInitStateIndex);
        codeStream.addDefinitelyAssignedVariables((Scope)currentScope, this.preConditionInitStateIndex);
      } 
      Constant cst = this.condition.optimizedBooleanConstant();
      boolean isConditionOptimizedFalse = (cst != Constant.NotAConstant && !cst.booleanValue());
      if (isConditionOptimizedFalse) {
        this.condition.generateCode(currentScope, codeStream, false);
      } else {
        this.condition.generateOptimizedBoolean(
            currentScope, 
            codeStream, 
            actionLabel, 
            null, 
            true);
      } 
    } 
    if (this.mergedInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
      codeStream.addDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
    } 
    if (this.breakLabel.forwardReferenceCount() > 0)
      this.breakLabel.place(); 
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }
  
  public StringBuffer printStatement(int indent, StringBuffer output) {
    printIndent(indent, output).append("do");
    if (this.action == null) {
      output.append(" ;\n");
    } else {
      output.append('\n');
      this.action.printStatement(indent + 1, output).append('\n');
    } 
    output.append("while (");
    return this.condition.printExpression(0, output).append(");");
  }
  
  public void resolve(BlockScope scope) {
    TypeBinding type = this.condition.resolveTypeExpecting(scope, (TypeBinding)TypeBinding.BOOLEAN);
    this.condition.computeConversion((Scope)scope, type, type);
    if (this.action != null)
      this.action.resolve(scope); 
  }
  
  public void traverse(ASTVisitor visitor, BlockScope scope) {
    if (visitor.visit(this, scope)) {
      if (this.action != null)
        this.action.traverse(visitor, scope); 
      this.condition.traverse(visitor, scope);
    } 
    visitor.endVisit(this, scope);
  }
  
  public boolean doesNotCompleteNormally() {
    Constant cst = this.condition.constant;
    boolean isConditionTrue = !(cst != null && (cst == Constant.NotAConstant || !cst.booleanValue()));
    cst = this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedTrue = (cst == null) ? true : ((cst != Constant.NotAConstant && cst.booleanValue()));
    if (isConditionTrue || isConditionOptimizedTrue)
      return !(this.action != null && this.action.breaksOut(null)); 
    if (this.action == null || this.action.breaksOut(null))
      return false; 
    return (this.action.doesNotCompleteNormally() && !this.action.completesByContinue());
  }
  
  public boolean completesByContinue() {
    return this.action.continuesAtOuterLabel();
  }
}
